@@ -4,9 +4,19 @@ import copy
 import inspect
 import json
 from collections import namedtuple
-from distutils.version import StrictVersion
 from functools import wraps
-from typing import Callable, Dict, Tuple, Union, Optional, List  # noqa
+from typing import (
+    Any,
+    Callable,
+    cast,
+    Dict,
+    List,
+    Optional,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+)
 from unittest.mock import Mock, patch
 from uuid import uuid4
 
@@ -19,6 +29,7 @@ from aiohttp import (
 )
 from aiohttp.helpers import TimerNoop
 from multidict import CIMultiDict, CIMultiDictProxy
+from pkg_resources import parse_version
 
 from .compat import (
     AIOHTTP_VERSION,
@@ -27,18 +38,22 @@ from .compat import (
     stream_reader_factory,
     merge_params,
     normalize_url,
+    RequestInfo,
 )
 
 
+_FuncT = TypeVar("_FuncT", bound=Callable[..., Any])
+
+
 class CallbackResult:
 
     def __init__(self, method: str = hdrs.METH_GET,
                  status: int = 200,
-                 body: str = '',
+                 body: Union[str, bytes] = '',
                  content_type: str = 'application/json',
-                 payload: Dict = None,
-                 headers: Dict = None,
-                 response_class: 'ClientResponse' = None,
+                 payload: Optional[Dict] = None,
+                 headers: Optional[Dict] = None,
+                 response_class: Optional[Type[ClientResponse]] = None,
                  reason: Optional[str] = None):
         self.method = method
         self.status = status
@@ -56,12 +71,12 @@ class RequestMatch(object):
     def __init__(self, url: Union[URL, str, Pattern],
                  method: str = hdrs.METH_GET,
                  status: int = 200,
-                 body: str = '',
-                 payload: Dict = None,
-                 exception: 'Exception' = None,
-                 headers: Dict = None,
+                 body: Union[str, bytes] = '',
+                 payload: Optional[Dict] = None,
+                 exception: Optional[Exception] = None,
+                 headers: Optional[Dict] = None,
                  content_type: str = 'application/json',
-                 response_class: 'ClientResponse' = None,
+                 response_class: Optional[Type[ClientResponse]] = None,
                  timeout: bool = False,
                  repeat: bool = False,
                  reason: Optional[str] = None,
@@ -95,7 +110,10 @@ class RequestMatch(object):
         return self.url_or_pattern == url
 
     def match_regexp(self, url: URL) -> bool:
-        return bool(self.url_or_pattern.match(str(url)))
+        # This method is used if and only if self.url_or_pattern is a pattern.
+        return bool(
+            self.url_or_pattern.match(str(url))  # type:ignore[union-attr]
+        )
 
     def match(self, method: str, url: URL) -> bool:
         if self.method != method.lower():
@@ -115,13 +133,13 @@ class RequestMatch(object):
 
     def _build_response(self, url: 'Union[URL, str]',
                         method: str = hdrs.METH_GET,
-                        request_headers: Dict = None,
+                        request_headers: Optional[Dict] = None,
                         status: int = 200,
-                        body: str = '',
+                        body: Union[str, bytes] = '',
                         content_type: str = 'application/json',
-                        payload: Dict = None,
-                        headers: Dict = None,
-                        response_class: 'ClientResponse' = None,
+                        payload: Optional[Dict] = None,
+                        headers: Optional[Dict] = None,
+                        response_class: Optional[Type[ClientResponse]] = None,
                         reason: Optional[str] = None) -> ClientResponse:
         if response_class is None:
             response_class = ClientResponse
@@ -131,12 +149,12 @@ class RequestMatch(object):
             body = str.encode(body)
         if request_headers is None:
             request_headers = {}
-        kwargs = {}
-        if AIOHTTP_VERSION >= StrictVersion('3.1.0'):
+        kwargs = {}  # type: Dict[str, Any]
+        if AIOHTTP_VERSION >= parse_version('3.1.0'):
             loop = Mock()
             loop.get_debug = Mock()
             loop.get_debug.return_value = True
-            kwargs['request_info'] = Mock(
+            kwargs['request_info'] = RequestInfo(
                 url=url,
                 method=method,
                 headers=CIMultiDictProxy(CIMultiDict(**request_headers)),
@@ -144,7 +162,7 @@ class RequestMatch(object):
             kwargs['writer'] = Mock()
             kwargs['continue100'] = None
             kwargs['timer'] = TimerNoop()
-            if AIOHTTP_VERSION < StrictVersion('3.3.0'):
+            if AIOHTTP_VERSION < parse_version('3.3.0'):
                 kwargs['auto_decompress'] = True
             kwargs['traces'] = []
             kwargs['loop'] = loop
@@ -161,7 +179,7 @@ class RequestMatch(object):
         for hdr in _headers.getall(hdrs.SET_COOKIE, ()):
             resp.cookies.load(hdr)
 
-        if AIOHTTP_VERSION >= StrictVersion('3.3.0'):
+        if AIOHTTP_VERSION >= parse_version('3.3.0'):
             # Reified attributes
             resp._headers = _headers
             resp._raw_headers = raw_headers
@@ -176,11 +194,8 @@ class RequestMatch(object):
         return resp
 
     async def build_response(
-        self, url: URL, **kwargs
+        self, url: URL, **kwargs: Any
     ) -> 'Union[ClientResponse, Exception]':
-        if self.exception is not None:
-            return self.exception
-
         if callable(self.callback):
             if asyncio.iscoroutinefunction(self.callback):
                 result = await self.callback(url, **kwargs)
@@ -188,6 +203,10 @@ class RequestMatch(object):
                 result = self.callback(url, **kwargs)
         else:
             result = None
+
+        if self.exception is not None:
+            return self.exception
+
         result = self if result is None else result
         resp = self._build_response(
             url=url,
@@ -212,7 +231,7 @@ class aioresponses(object):
     _responses = None  # type: List[ClientResponse]
     requests = None  # type: Dict
 
-    def __init__(self, **kwargs):
+    def __init__(self, **kwargs: Any):
         self._param = kwargs.pop('param', None)
         self._passthrough = kwargs.pop('passthrough', [])
         self.patcher = patch('aiohttp.client.ClientSession._request',
@@ -224,10 +243,10 @@ class aioresponses(object):
         self.start()
         return self
 
-    def __exit__(self, exc_type, exc_val, exc_tb):
+    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
         self.stop()
 
-    def __call__(self, f):
+    def __call__(self, f: _FuncT) -> _FuncT:
         def _pack_arguments(ctx, *args, **kwargs) -> Tuple[Tuple, Dict]:
             if self._param:
                 kwargs[self._param] = ctx
@@ -247,13 +266,13 @@ class aioresponses(object):
                 with self as ctx:
                     args, kwargs = _pack_arguments(ctx, *args, **kwargs)
                     return f(*args, **kwargs)
-        return wrapped
+        return cast(_FuncT, wrapped)
 
-    def clear(self):
+    def clear(self) -> None:
         self._responses.clear()
         self._matches.clear()
 
-    def start(self):
+    def start(self) -> None:
         self._responses = []
         self._matches = {}
         self.patcher.start()
@@ -265,35 +284,35 @@ class aioresponses(object):
         self.patcher.stop()
         self.clear()
 
-    def head(self, url: 'Union[URL, str, Pattern]', **kwargs):
+    def head(self, url: 'Union[URL, str, Pattern]', **kwargs: Any) -> None:
         self.add(url, method=hdrs.METH_HEAD, **kwargs)
 
-    def get(self, url: 'Union[URL, str, Pattern]', **kwargs):
+    def get(self, url: 'Union[URL, str, Pattern]', **kwargs: Any) -> None:
         self.add(url, method=hdrs.METH_GET, **kwargs)
 
-    def post(self, url: 'Union[URL, str, Pattern]', **kwargs):
+    def post(self, url: 'Union[URL, str, Pattern]', **kwargs: Any) -> None:
         self.add(url, method=hdrs.METH_POST, **kwargs)
 
-    def put(self, url: 'Union[URL, str, Pattern]', **kwargs):
+    def put(self, url: 'Union[URL, str, Pattern]', **kwargs: Any) -> None:
         self.add(url, method=hdrs.METH_PUT, **kwargs)
 
-    def patch(self, url: 'Union[URL, str, Pattern]', **kwargs):
+    def patch(self, url: 'Union[URL, str, Pattern]', **kwargs: Any) -> None:
         self.add(url, method=hdrs.METH_PATCH, **kwargs)
 
-    def delete(self, url: 'Union[URL, str, Pattern]', **kwargs):
+    def delete(self, url: 'Union[URL, str, Pattern]', **kwargs: Any) -> None:
         self.add(url, method=hdrs.METH_DELETE, **kwargs)
 
-    def options(self, url: 'Union[URL, str, Pattern]', **kwargs):
+    def options(self, url: 'Union[URL, str, Pattern]', **kwargs: Any) -> None:
         self.add(url, method=hdrs.METH_OPTIONS, **kwargs)
 
     def add(self, url: 'Union[URL, str, Pattern]', method: str = hdrs.METH_GET,
             status: int = 200,
-            body: str = '',
-            exception: 'Exception' = None,
+            body: Union[str, bytes] = '',
+            exception: Optional[Exception] = None,
             content_type: str = 'application/json',
-            payload: Dict = None,
-            headers: Dict = None,
-            response_class: 'ClientResponse' = None,
+            payload: Optional[Dict] = None,
+            headers: Optional[Dict] = None,
+            response_class: Optional[Type[ClientResponse]] = None,
             repeat: bool = False,
             timeout: bool = False,
             reason: Optional[str] = None,
@@ -315,6 +334,113 @@ class aioresponses(object):
             callback=callback,
         ))
 
+    def _format_call_signature(self, *args, **kwargs) -> str:
+        message = '%s(%%s)' % self.__class__.__name__ or 'mock'
+        formatted_args = ''
+        args_string = ', '.join([repr(arg) for arg in args])
+        kwargs_string = ', '.join([
+            '%s=%r' % (key, value) for key, value in kwargs.items()
+        ])
+        if args_string:
+            formatted_args = args_string
+        if kwargs_string:
+            if formatted_args:
+                formatted_args += ', '
+            formatted_args += kwargs_string
+
+        return message % formatted_args
+
+    def assert_not_called(self):
+        """assert that the mock was never called.
+        """
+        if len(self.requests) != 0:
+            msg = ("Expected '%s' to not have been called. Called %s times."
+                   % (self.__class__.__name__,
+                      len(self._responses)))
+            raise AssertionError(msg)
+
+    def assert_called(self):
+        """assert that the mock was called at least once.
+        """
+        if len(self.requests) == 0:
+            msg = ("Expected '%s' to have been called."
+                   % (self.__class__.__name__,))
+            raise AssertionError(msg)
+
+    def assert_called_once(self):
+        """assert that the mock was called only once.
+        """
+        call_count = len(self.requests)
+        if call_count == 1:
+            call_count = len(list(self.requests.values())[0])
+        if not call_count == 1:
+            msg = ("Expected '%s' to have been called once. Called %s times."
+                   % (self.__class__.__name__,
+                      call_count))
+
+            raise AssertionError(msg)
+
+    def assert_called_with(self, url: 'Union[URL, str, Pattern]',
+                           method: str = hdrs.METH_GET,
+                           *args: Any,
+                           **kwargs: Any):
+        """assert that the last call was made with the specified arguments.
+
+        Raises an AssertionError if the args and keyword args passed in are
+        different to the last call to the mock."""
+        url = normalize_url(merge_params(url, kwargs.get('params')))
+        method = method.upper()
+        key = (method, url)
+        try:
+            expected = self.requests[key][-1]
+        except KeyError:
+            expected_string = self._format_call_signature(
+                url, method=method, *args, **kwargs
+            )
+            raise AssertionError(
+                '%s call not found' % expected_string
+            )
+        actual = self._build_request_call(method, *args, **kwargs)
+        if not expected == actual:
+            expected_string = self._format_call_signature(
+                expected,
+            )
+            actual_string = self._format_call_signature(
+                actual
+            )
+            raise AssertionError(
+                '%s != %s' % (expected_string, actual_string)
+            )
+
+    def assert_any_call(self, url: 'Union[URL, str, Pattern]',
+                        method: str = hdrs.METH_GET,
+                        *args: Any,
+                        **kwargs: Any):
+        """assert the mock has been called with the specified arguments.
+        The assert passes if the mock has *ever* been called, unlike
+        `assert_called_with` and `assert_called_once_with` that only pass if
+        the call is the most recent one."""
+        url = normalize_url(merge_params(url, kwargs.get('params')))
+        method = method.upper()
+        key = (method, url)
+
+        try:
+            self.requests[key]
+        except KeyError:
+            expected_string = self._format_call_signature(
+                url, method=method, *args, **kwargs
+            )
+            raise AssertionError(
+                '%s call not found' % expected_string
+            )
+
+    def assert_called_once_with(self, *args: Any, **kwargs: Any):
+        """assert that the mock was called once with the specified arguments.
+        Raises an AssertionError if the args and keyword args passed in are
+        different to the only call to the mock."""
+        self.assert_called_once()
+        self.assert_called_with(*args, **kwargs)
+
     @staticmethod
     def is_exception(resp_or_exc: Union[ClientResponse, Exception]) -> bool:
         if inspect.isclass(resp_or_exc):
@@ -327,8 +453,10 @@ class aioresponses(object):
         return False
 
     async def match(
-        self, method: str, url: URL,
-        allow_redirects: bool = True, **kwargs: Dict
+        self, method: str,
+        url: URL,
+        allow_redirects: bool = True,
+        **kwargs: Any
     ) -> Optional['ClientResponse']:
         history = []
         while True:
@@ -346,25 +474,32 @@ class aioresponses(object):
 
             if self.is_exception(response_or_exc):
                 raise response_or_exc
-            is_redirect = response_or_exc.status in (301, 302, 303, 307, 308)
+            # If response_or_exc was an exception, it would have been raised.
+            # At this point we can be sure it's a ClientResponse
+            response: ClientResponse
+            response = response_or_exc  # type:ignore[assignment]
+            is_redirect = response.status in (301, 302, 303, 307, 308)
             if is_redirect and allow_redirects:
-                if hdrs.LOCATION not in response_or_exc.headers:
+                if hdrs.LOCATION not in response.headers:
                     break
-                history.append(response_or_exc)
-                url = URL(response_or_exc.headers[hdrs.LOCATION])
+                history.append(response)
+                redirect_url = URL(response.headers[hdrs.LOCATION])
+                if redirect_url.is_absolute():
+                    url = redirect_url
+                else:
+                    url = url.join(redirect_url)
                 method = 'get'
                 continue
             else:
                 break
 
-        response_or_exc._history = tuple(history)
-
-        return response_or_exc
+        response._history = tuple(history)
+        return response
 
     async def _request_mock(self, orig_self: ClientSession,
                             method: str, url: 'Union[URL, str]',
                             *args: Tuple,
-                            **kwargs: Dict) -> 'ClientResponse':
+                            **kwargs: Any) -> 'ClientResponse':
         """Return mocked response object or raise connection error."""
         if orig_self.closed:
             raise RuntimeError('Session is closed')
@@ -380,12 +515,8 @@ class aioresponses(object):
 
         key = (method, url)
         self.requests.setdefault(key, [])
-        try:
-            kwargs_copy = copy.deepcopy(kwargs)
-        except TypeError:
-            # Handle the fact that some values cannot be deep copied
-            kwargs_copy = kwargs
-        self.requests[key].append(RequestCall(args, kwargs_copy))
+        request_call = self._build_request_call(method, *args, **kwargs)
+        self.requests[key].append(request_call)
 
         response = await self.match(method, url, **kwargs)
 
@@ -409,3 +540,19 @@ class aioresponses(object):
             response.raise_for_status()
 
         return response
+
+    def _build_request_call(self, method: str = hdrs.METH_GET,
+                            *args: Any,
+                            allow_redirects: bool = True,
+                            **kwargs: Any):
+        """Return request call."""
+        kwargs.setdefault('allow_redirects', allow_redirects)
+        if method == 'POST':
+            kwargs.setdefault('data', None)
+
+        try:
+            kwargs_copy = copy.deepcopy(kwargs)
+        except (TypeError, ValueError):
+            # Handle the fact that some values cannot be deep copied
+            kwargs_copy = kwargs
+        return RequestCall(args, kwargs_copy)
