@@ -143,6 +143,9 @@ class AIOResponsesTestCase(AsyncTestCase):
         )
         self.assertIsInstance(response, ClientResponse)
         self.assertEqual(response.status, 200)
+        self.assertEqual(len(m.requests), 2)
+        with self.assertRaises(AssertionError):
+            m.assert_called_once()
 
     @aioresponses()
     def test_method_dont_match(self, m):
@@ -150,6 +153,59 @@ class AIOResponsesTestCase(AsyncTestCase):
         with self.assertRaises(ClientConnectionError):
             self.run_async(self.session.post(self.url))
 
+    @aioresponses()
+    def test_post_with_data(self, m: aioresponses):
+        body = {'foo': 'bar'}
+        payload = {'spam': 'eggs'}
+        user_agent = {'User-Agent': 'aioresponses'}
+        m.post(
+            self.url,
+            payload=payload,
+            headers=dict(connection='keep-alive'),
+            body=body,
+        )
+        response = self.run_async(
+            self.session.post(
+                self.url,
+                data=payload,
+                headers=user_agent
+            )
+        )
+        self.assertIsInstance(response, ClientResponse)
+        self.assertEqual(response.status, 200)
+        response_data = self.run_async(response.json())
+        self.assertEqual(response_data, payload)
+        m.assert_called_once_with(
+            self.url,
+            method='POST',
+            data=payload,
+            headers={'User-Agent': 'aioresponses'}
+        )
+        # Wrong data
+        with self.assertRaises(AssertionError):
+            m.assert_called_once_with(
+                self.url,
+                method='POST',
+                data=body,
+                headers={'User-Agent': 'aioresponses'}
+            )
+        # Wrong url
+        with self.assertRaises(AssertionError):
+            m.assert_called_once_with(
+                'http://httpbin.org/',
+                method='POST',
+                data=payload,
+                headers={'User-Agent': 'aioresponses'}
+            )
+        # Wrong headers
+        with self.assertRaises(AssertionError):
+            m.assert_called_once_with(
+                self.url,
+                method='POST',
+                data=payload,
+                headers={'User-Agent': 'aiorequest'}
+            )
+
     @aioresponses()
     async def test_streaming(self, m):
         m.get(self.url, body='Test')
@@ -492,6 +548,60 @@ class AIOResponsesTestCase(AsyncTestCase):
         data = self.run_async(response.read())
         assert data == body
 
+    @aioresponses()
+    def test_assert_not_called(self, m: aioresponses):
+        m.get(self.url)
+        m.assert_not_called()
+        self.run_async(self.session.get(self.url))
+        with self.assertRaises(AssertionError):
+            m.assert_not_called()
+
+    @aioresponses()
+    def test_assert_called(self, m: aioresponses):
+        m.get(self.url)
+        with self.assertRaises(AssertionError):
+            m.assert_called()
+        self.run_async(self.session.get(self.url))
+
+        m.assert_called_once()
+        m.assert_called_once_with(self.url)
+        m.assert_called_with(self.url)
+        with self.assertRaises(AssertionError):
+            m.assert_not_called()
+
+        with self.assertRaises(AssertionError):
+            m.assert_called_with("http://foo.bar")
+
+    @aioresponses()
+    async def test_assert_called_twice(self, m: aioresponses):
+        m.get(self.url, repeat=True)
+        m.assert_not_called()
+        await self.session.get(self.url)
+        await self.session.get(self.url)
+        with self.assertRaises(AssertionError):
+            m.assert_called_once()
+
+    @aioresponses()
+    async def test_assert_any_call(self, m: aioresponses):
+        http_bin_url = "http://httpbin.org"
+        m.get(self.url)
+        m.get(http_bin_url)
+        await self.session.get(self.url)
+        response = await self.session.get(http_bin_url)
+        self.assertEqual(response.status, 200)
+        m.assert_any_call(self.url)
+        m.assert_any_call(http_bin_url)
+
+    @aioresponses()
+    async def test_assert_any_call_not_called(self, m: aioresponses):
+        http_bin_url = "http://httpbin.org"
+        m.get(self.url)
+        response = await self.session.get(self.url)
+        self.assertEqual(response.status, 200)
+        m.assert_any_call(self.url)
+        with self.assertRaises(AssertionError):
+            m.assert_any_call(http_bin_url)
+
     @aioresponses()
     async def test_exception_requests_are_tracked(self, mocked):
         kwargs = {"json": [42], "allow_redirects": True}
@@ -655,3 +765,21 @@ class AIOResponseRedirectTest(AsyncTestCase):
         request_info = response.request_info
         assert str(request_info.url) == self.url
         assert request_info.headers == headers
+
+    @aioresponses()
+    async def test_relative_url_redirect_followed(self, rsps):
+        base_url = "https://httpbin.org"
+        url = f"{base_url}/foo/bar"
+        rsps.get(
+            url,
+            status=307,
+            headers={"Location": "../baz"},
+        )
+        rsps.get(f"{base_url}/baz")
+
+        response = await self.session.get(url, allow_redirects=True)
+
+        self.assertEqual(response.status, 200)
+        self.assertEqual(str(response.url), f"{base_url}/baz")
+        self.assertEqual(len(response.history), 1)
+        self.assertEqual(str(response.history[0].url), url)
